#include "TecModule.h"
#include "TecParameters.h"

/**********************************************************************************************************************/

TecModule::TecModule(EntityOwner* owner, const std::string& name, const std::string& description,
    ctk::HierarchyModifier hierarchyModifier, const std::unordered_set<std::string>& tags)
: ctk::ApplicationModule(owner, name, description, hierarchyModifier, tags) {
  // example how to fill the map
    writeIntMap.emplace(
        TecFrame::TecParameter::ExpFanControlEnable, ChimeraTK::ScalarPollInput<int>{this, "ctrl_ch1_fanControlEnable_wr", "unit", "description"}); 
      
    writeFloatMap.emplace(
        TecFrame::TecParameter::TemTargetObjectTemp, ChimeraTK::ScalarPollInput<float>{this, "ctrl_ch1_targetObjectTemp_wr", "unit", "description"});        
        
    readIntMap.emplace(
        TecFrame::TecParameter::ExpFanControlEnable, ChimeraTK::ScalarOutput<int>{this, "ctrl_ch1_fanControlEnable_rd", "unit", "description"});       
    readIntMap.emplace(
        TecFrame::TecParameter::MonSerialNumber, ChimeraTK::ScalarOutput<int>{this, "mon_serialNumber_rd", "unit", "description"});   
           
    readFloatMap.emplace(
        TecFrame::TecParameter::TemTargetObjectTemp, ChimeraTK::ScalarOutput<float>{this, "ctrl_ch1_targetObjectTemp_rd", "unit", "description"});         
    readFloatMap.emplace(
        TecFrame::TecParameter::MonSinkTemperature, ChimeraTK::ScalarOutput<float>{this, "mon_ch1_SinkTemp_rd", "unit", "description"});      
        
        
      
      
  // Note: The maps writeGroupMap, writeGroupToFlashMap, readGroupMap need to be filled in the mainLoop(). The
  // corresponding ScalarPushInputs need to be created here and placed onto the groupCommandMap.

  // example how to fill a group command -> this part needs to later automatically generated from std::list<ParameterDescriptor> descriptors
  // there is ticket to automate it
  // for now manually entries prepared for start of the Focus Week
  
  // commands for read-only for channel 1 parameters in group 1 where TecFrame::TecParameter::MonSinkTemperature query includes
  groupCommandMap.emplace("readCh1Group1",
      ChimeraTK::ScalarPushInput<int>{this, "readCh1Group1_wr", "", "Trigger read of channel 1 read-only parameters group 1"});
  // read-only for generic parameter in group 1 for TecFrame::TecParameter::MonSerialNumber query includes
  groupCommandMap.emplace("readGroup1",
      ChimeraTK::ScalarPushInput<int>{this, "readGroup1_wr", "", "Trigger read of generic read-only parameters group 1."});
  
  // commands for read-write group 7 for channel 1 where TecFrame::TecParameter::ExpFanControlEnable query includes
  groupCommandMap.emplace("readCh1WriteGroup7",
      ChimeraTK::ScalarPushInput<int>{this, "readCh1WriteGroup7_wr", "", "Trigger read of channel 1 read-write group 7"});
  groupCommandMap.emplace("writeCh1Group7",
      ChimeraTK::ScalarPushInput<int>{this, "writeCh1Group7_wr", "", "Trigger write of channel 1 read-write group 7"});
  groupCommandMap.emplace("writeToFlashCh1Group7",
      ChimeraTK::ScalarPushInput<int>{this, "writeToFlashCh1Group7_wr", "", "Trigger write to FLASH of channel 1 read-write group 7"});
  groupCommandMap.emplace("setRbCh1Group7",
      ChimeraTK::ScalarPushInput<int>{this, "setRbCh1Group7_wr", "", "Trigger set from readback for channel 1 read-write group 7"});      
      
  // commands for read-write group 1 for channel 1 where TecFrame::TecParameter::TemTargetObjectTemp query includes
  groupCommandMap.emplace("readCh1WriteGroup1",
      ChimeraTK::ScalarPushInput<int>{this, "readCh1WriteGroup1_wr", "", "Trigger read of channel 1 read-write group 1"});
  groupCommandMap.emplace("writeCh1Group1",
      ChimeraTK::ScalarPushInput<int>{this, "writeCh1Group1_wr", "", "Trigger write of channel 1 read-write group 1"});
  groupCommandMap.emplace("writeToFlashCh1Group1",
      ChimeraTK::ScalarPushInput<int>{this, "writeToFlashCh1Group1_wr", "", "Trigger write to FLASH of channel 1 read-write group 1"});
  groupCommandMap.emplace("setRbCh1Group1",
      ChimeraTK::ScalarPushInput<int>{this, "setRbCh1Group1_wr", "", "Trigger set from readback for channel 1 read-write group 1"});       
}

/**********************************************************************************************************************/

void TecModule::mainLoop() {
  std::cout << "Processing TecModule::mainLoop()" << std::endl;

  tecIface = std::make_shared<TecChimeraInterface>(
      triggerCommands, responseStatus, commandMemoryRaw_wr, commandMemoryRaw_rd, true);
  tec = TecController(tecIface);

  /* This code is just for demonstration/testing purposes. We can use code like this to revese engineer the teclib
   * when implementing the TecDummy backend. It will be removed later! */
  tec.addFrame(0, TecFrame::TecParameter::ObjMaxTempChange, TecFrame::TecFrameType::TecParameterValueSet, 1, 0.0F);
  tec.addFrame(0, TecFrame::TecParameter::DeviceStatus, TecFrame::TecFrameType::TecParameterValueRead, 1);
  std::cout << "constructAllFrames:" << std::endl;
  tec.constructAllFrames();
  for(uint8_t i = 0; i < tec.getFrameCount(); i++) {
    tec.getFrameAt(i)->print();
  }
  std::cout << "copyCommandsToHardware:" << std::endl;
  tec.copyCommandsToHardware();
  std::cout << "triggerHardwareCommands:" << std::endl;
  tec.triggerHardwareCommands();
  std::cout << "readHardwareResponses:" << std::endl;
  auto ret = tec.readHardwareResponses();
  std::cout << " -> return " << ret << std::endl;
  std::cout << "copyResponsesFromHardware:" << std::endl;
  tec.copyResponsesFromHardware();
  std::cout << "parseAllFrames:" << std::endl;
  tec.parseAllFrames();
  for(uint8_t i = 0; i < tec.getFrameCount(); i++) {
    std::cout << "STATE = " << tec.getFrameAt(i)->getState() << std::endl;
    tec.getFrameAt(i)->print();
  }
  std::cout << "done." << std::endl;

    // These the group maps should to be filled by iterating the descriptor list and obtaining the corresponding inputs from
    // the groupCommandMap. For now it is done manually but there is a tictet to automate it. 
    readGroupMap.emplace(groupCommandMap.at("readCh1Group1").getId(),
            std::list<TecFrame::TecParameter>{TecFrame::TecParameter::MonSinkTemperature});
    readGroupMap.emplace(groupCommandMap.at("readGroup1").getId(),
            std::list<TecFrame::TecParameter>{TecFrame::TecParameter::MonSerialNumber});
    readGroupMap.emplace(groupCommandMap.at("readCh1WriteGroup7").getId(),
            std::list<TecFrame::TecParameter>{TecFrame::TecParameter::ExpFanControlEnable});
    readGroupMap.emplace(groupCommandMap.at("readCh1WriteGroup1").getId(),
            std::list<TecFrame::TecParameter>{TecFrame::TecParameter::TemTargetObjectTemp});

    writeGroupMap.emplace(groupCommandMap.at("writeCh1Group7").getId(),
            std::list<TecFrame::TecParameter>{TecFrame::TecParameter::ExpFanControlEnable});
    writeGroupMap.emplace(groupCommandMap.at("writeCh1Group1").getId(),
            std::list<TecFrame::TecParameter>{TecFrame::TecParameter::TemTargetObjectTemp});

    writeGroupToFlashMap.emplace(groupCommandMap.at("writeToFlashCh1Group7").getId(),
            std::list<TecFrame::TecParameter>{TecFrame::TecParameter::ExpFanControlEnable});
    writeGroupToFlashMap.emplace(groupCommandMap.at("writeToFlashCh1Group7").getId(),
            std::list<TecFrame::TecParameter>{TecFrame::TecParameter::TemTargetObjectTemp});           
              
      
      
  auto group = readAnyGroup();
  while(true) {
    auto id = group.readAny();
    if(id == periodicRead.trigger.getId()) {
      readPeriodic();
    }
    else if(readGroupMap.find(id) != readGroupMap.end()) {
      readGroup(readGroupMap.at(id));
    }
    else if(writeGroupMap.find(id) != writeGroupMap.end()) {
      writeGroup(writeGroupMap.at(id));
    }
    else if(writeGroupToFlashMap.find(id) != writeGroupToFlashMap.end()) {
      writeGroupToFlash(writeGroupToFlashMap.at(id));
    }
    else {
      std::cout << "Unexpected input (id = " << id << ") has been written, don't know what to do!" << std::endl;
    }
  }
}

/**********************************************************************************************************************/

void TecModule::readPeriodic() {
  std::cout << "readPeroidoc()" << std::endl;
}

/**********************************************************************************************************************/

void TecModule::readGroup(std::list<TecFrame::TecParameter>& group) {
  std::cout << "readGroup()" << std::endl;
  for(auto& param : group) std::cout << "Read param: " << param << std::endl;
}

/**********************************************************************************************************************/

void TecModule::writeGroup(std::list<TecFrame::TecParameter>& group) {
  std::cout << "writeGroup()" << std::endl;
  for(auto& param : group) std::cout << "Write param: " << param << std::endl;
}

/**********************************************************************************************************************/

void TecModule::writeGroupToFlash(std::list<TecFrame::TecParameter>& group) {
  std::cout << "writeGroupToFlash()" << std::endl;
  for(auto& param : group) std::cout << "Write param to flash: " << param << std::endl;
}

/**********************************************************************************************************************/
